from typing import Optional

import pytest

from sanic_ext.extensions.openapi.types import Definition


@pytest.fixture
def Thing():
    class Thing(Definition):
        name: str
        foo: Optional[bool]
        bar: Optional[bool]

        def __init__(self, name, foo=None, bar=None):
            super().__init__(name=name, foo=foo, bar=bar)

    return Thing


def test_serialize_null_show_by_default(Thing):
    thing = Thing(name="ok")
    serialized = thing.serialize()

    assert serialized["name"] == "ok"
    assert serialized["foo"] is None
    assert serialized["bar"] is None


def test_serialize_some_nullable(Thing):
    Thing.__nullable__ = ["foo"]
    thing = Thing(name="ok")
    serialized = thing.serialize()

    assert serialized["name"] == "ok"
    assert serialized["foo"] is None
    assert "bar" not in serialized


def test_serialize_no_nullable(Thing):
    Thing.__nullable__ = False
    thing = Thing(name="ok")
    serialized = thing.serialize()

    assert serialized["name"] == "ok"
    assert "foo" not in serialized
    assert "bar" not in serialized



from sanic_ext.extensions.openapi.builders import SpecificationBuilder


def test_specification_singleton():
    spec = SpecificationBuilder()
    spec.tag("foo")

    new_spec = SpecificationBuilder()

    assert spec is new_spec
    assert spec.tags == new_spec.tags


def test_specification_reset(get_docs):
    spec = SpecificationBuilder()

    docs = get_docs()
    assert len(docs["tags"]) == 0

    spec.tag("foo")
    docs = get_docs()
    assert len(docs["tags"]) == 1

    spec.reset()
    docs = get_docs()
    assert len(docs["tags"]) == 0



from __future__ import annotations

import os
from typing import Any, Dict, List, Optional, Sequence, Union

from sanic import Sanic
from sanic.config import Config as SanicConfig
from sanic.exceptions import SanicException
from sanic.signals import Event


class Config(SanicConfig):
    def __init__(
        self,
        cors: bool = True,
        cors_allow_headers: str = "*",
        cors_always_send: bool = True,
        cors_automatic_options: bool = True,
        cors_expose_headers: str = "",
        cors_max_age: int = 5,
        cors_methods: str = "",
        cors_origins: Union[str, List[str]] = "",
        cors_send_wildcard: bool = False,
        cors_supports_credentials: bool = False,
        cors_vary_header: bool = True,
        health: bool = False,
        health_endpoint: bool = False,
        health_max_misses: int = 3,
        health_missed_threshhold: int = 10,
        health_monitor: bool = True,
        health_report_interval: int = 5,
        health_uri_to_info: str = "",
        health_url_prefix: str = "/__health__",
        http_all_methods: bool = True,
        http_auto_head: bool = True,
        http_auto_options: bool = True,
        http_auto_trace: bool = False,
        injection_signal: Union[str, Event] = Event.HTTP_ROUTING_AFTER,
        logging: bool = False,
        logging_queue_max_size: int = 4096,
        oas: bool = True,
        oas_autodoc: bool = True,
        oas_ignore_head: bool = True,
        oas_ignore_options: bool = True,
        oas_path_to_redoc_html: Optional[str] = None,
        oas_path_to_swagger_html: Optional[str] = None,
        oas_ui_default: Optional[str] = "redoc",
        oas_ui_redoc: bool = True,
        oas_ui_swagger: bool = True,
        oas_ui_swagger_version: str = "4.10.3",
        oas_uri_to_config: str = "/swagger-config",
        oas_uri_to_json: str = "/openapi.json",
        oas_uri_to_redoc: str = "/redoc",
        oas_uri_to_swagger: str = "/swagger",
        oas_url_prefix: str = "/docs",
        swagger_ui_configuration: Optional[Dict[str, Any]] = None,
        templating_path_to_templates: Union[
            str, os.PathLike, Sequence[Union[str, os.PathLike]]
        ] = "templates",
        templating_enable_async: bool = True,
        trace_excluded_headers: Sequence[str] = ("authorization", "cookie"),
        **kwargs,
    ):
        self.CORS = cors
        self.CORS_ALLOW_HEADERS = cors_allow_headers
        self.CORS_ALWAYS_SEND = cors_always_send
        self.CORS_AUTOMATIC_OPTIONS = cors_automatic_options
        self.CORS_EXPOSE_HEADERS = cors_expose_headers
        self.CORS_MAX_AGE = cors_max_age
        self.CORS_METHODS = cors_methods
        self.CORS_ORIGINS = cors_origins
        self.CORS_SEND_WILDCARD = cors_send_wildcard
        self.CORS_SUPPORTS_CREDENTIALS = cors_supports_credentials
        self.CORS_VARY_HEADER = cors_vary_header
        self.HEALTH = health
        self.HEALTH_ENDPOINT = health_endpoint
        self.HEALTH_MAX_MISSES = health_max_misses
        self.HEALTH_MISSED_THRESHHOLD = health_missed_threshhold
        self.HEALTH_MONITOR = health_monitor
        self.HEALTH_REPORT_INTERVAL = health_report_interval
        self.HEALTH_URI_TO_INFO = health_uri_to_info
        self.HEALTH_URL_PREFIX = health_url_prefix
        self.HTTP_ALL_METHODS = http_all_methods
        self.HTTP_AUTO_HEAD = http_auto_head
        self.HTTP_AUTO_OPTIONS = http_auto_options
        self.HTTP_AUTO_TRACE = http_auto_trace
        self.INJECTION_SIGNAL = injection_signal
        self.LOGGING = logging
        self.LOGGING_QUEUE_MAX_SIZE = logging_queue_max_size
        self.OAS = oas
        self.OAS_AUTODOC = oas_autodoc
        self.OAS_IGNORE_HEAD = oas_ignore_head
        self.OAS_IGNORE_OPTIONS = oas_ignore_options
        self.OAS_PATH_TO_REDOC_HTML = oas_path_to_redoc_html
        self.OAS_PATH_TO_SWAGGER_HTML = oas_path_to_swagger_html
        self.OAS_UI_DEFAULT = oas_ui_default
        self.OAS_UI_REDOC = oas_ui_redoc
        self.OAS_UI_SWAGGER = oas_ui_swagger
        self.OAS_UI_SWAGGER_VERSION = oas_ui_swagger_version
        self.OAS_URI_TO_CONFIG = oas_uri_to_config
        self.OAS_URI_TO_JSON = oas_uri_to_json
        self.OAS_URI_TO_REDOC = oas_uri_to_redoc
        self.OAS_URI_TO_SWAGGER = oas_uri_to_swagger
        self.OAS_URL_PREFIX = oas_url_prefix
        self.SWAGGER_UI_CONFIGURATION = swagger_ui_configuration or {
            "apisSorter": "alpha",
            "operationsSorter": "alpha",
            "docExpansion": "full",
        }
        self.TEMPLATING_PATH_TO_TEMPLATES = templating_path_to_templates
        self.TEMPLATING_ENABLE_ASYNC = templating_enable_async
        self.TRACE_EXCLUDED_HEADERS = trace_excluded_headers

        if isinstance(self.TRACE_EXCLUDED_HEADERS, str):
            self.TRACE_EXCLUDED_HEADERS = tuple(
                self.TRACE_EXCLUDED_HEADERS.split(",")
            )

        if isinstance(self.INJECTION_SIGNAL, str):
            self.INJECTION_SIGNAL = Event(self.INJECTION_SIGNAL)

        valid_signals = ("http.handler.before", "http.routing.after")
        if self.INJECTION_SIGNAL.value not in valid_signals:
            raise SanicException(
                f"Injection signal may only be one of {valid_signals}"
            )

        self.load({key.upper(): value for key, value in kwargs.items()})

    @classmethod
    def from_dict(cls, mapping) -> Config:
        return cls(**mapping)


def add_fallback_config(
    app: Sanic, config: Optional[Config] = None, **kwargs
) -> Config:
    if config is None:
        config = Config(**kwargs)

    app.config.update(
        {key: value for key, value in config.items() if key not in app.config}
    )
    config.update(
        {
            key: value
            for key, value in app.config.items()
            if key in config and value != config.get(key)
        }
    )

    return config



from typing import Any, Dict

from sanic import Sanic


def get_spec(app: Sanic) -> Dict[str, Any]:
    test_client = app.test_client
    _, response = test_client.get("/docs/openapi.json")
    return response.json


def get_path(app: Sanic, path: str, method: str = "get") -> Dict[str, Any]:
    spec = get_spec(app)
    return spec["paths"][path][method.lower()]



from sanic import Blueprint, Request, Sanic
from sanic.response import json
from sanic.worker.inspector import Inspector


def setup_health_endpoint(app: Sanic) -> None:
    bp = Blueprint("SanicHealth", url_prefix=app.config.HEALTH_URL_PREFIX)

    @bp.get(app.config.HEALTH_URI_TO_INFO)
    async def info(request: Request):
        return json(Inspector.make_safe(dict(request.app.m.workers)))

    app.blueprint(bp)



from typing import Any, Callable, Dict, Optional, Tuple, Type

from .constructor import Constructor


class InjectionRegistry:
    def __init__(self):
        self._registry: Dict[Type, Optional[Callable[..., Any]]] = {}

    def __getitem__(self, key):
        return self._registry[key]

    def __str__(self) -> str:
        return str(self._registry)

    def __contains__(self, other: Any):
        return other in self._registry

    def get(self, key, default=None):
        return self._registry.get(key, default)

    def register(
        self, _type: Type, constructor: Optional[Callable[..., Any]]
    ) -> None:
        constructor = constructor or _type
        constructor = Constructor(constructor)
        self._registry[_type] = constructor

    def finalize(self, allowed_types):
        for constructor in self._registry.values():
            if isinstance(constructor, Constructor):
                constructor.prepare(self, allowed_types)

    @property
    def length(self):
        return len(self._registry)


class SignatureRegistry:
    def __init__(self):
        self._registry: Dict[
            str, Dict[str, Tuple[Type, Optional[Callable[..., Any]]]]
        ] = {}

    def __getitem__(self, key):
        return self._registry[key]

    def __str__(self) -> str:
        return str(self._registry)

    def get(self, key, default=None):
        return self._registry.get(key, default)

    def register(
        self,
        route_name: str,
        injections: Dict[str, Tuple[Type, Optional[Callable[..., Any]]]],
    ) -> None:
        self._registry[route_name] = injections



import re

import pytest
from sanic import Sanic

from sanic_ext import Extend, Extension
from sanic_ext.extensions.http.extension import HTTPExtension
from sanic_ext.extensions.injection.extension import InjectionExtension
from sanic_ext.extensions.openapi.builders import (
    OperationStore,
    SpecificationBuilder,
)
from sanic_ext.extensions.openapi.extension import OpenAPIExtension

slugify = re.compile(r"[^a-zA-Z0-9_\-]")


@pytest.fixture(autouse=True)
def reset_globals():
    yield
    SpecificationBuilder.reset()
    OperationStore.reset()
    Extend.reset()
    Extension.reset()
    Sanic._app_registry.clear()


@pytest.fixture(autouse=True)
def reset_extensions():
    yield
    for ext in (HTTPExtension, InjectionExtension, OpenAPIExtension):
        ext._singleton = None


@pytest.fixture
def bare_app(request):
    app = Sanic(slugify.sub("-", request.node.name))

    yield app


@pytest.fixture
def app(bare_app):
    Extend(bare_app)

    yield bare_app


@pytest.fixture
def get_docs(app):
    def fetch():
        nonlocal app

        _, response = app.test_client.get("/docs/openapi.json")
        return response.json

    return fetch



from __future__ import annotations

from string import ascii_lowercase
from types import SimpleNamespace
from typing import Any, Callable, Dict, List, Optional, Type, Union
from warnings import warn

from sanic import Sanic, __version__
from sanic.exceptions import SanicException
from sanic.log import logger

from sanic_ext.config import Config, add_fallback_config
from sanic_ext.extensions.base import Extension
from sanic_ext.extensions.health.extension import HealthExtension
from sanic_ext.extensions.http.extension import HTTPExtension
from sanic_ext.extensions.injection.extension import InjectionExtension
from sanic_ext.extensions.injection.registry import InjectionRegistry
from sanic_ext.extensions.logging.extension import LoggingExtension
from sanic_ext.extensions.openapi.builders import SpecificationBuilder
from sanic_ext.extensions.openapi.extension import OpenAPIExtension
from sanic_ext.utils.string import camel_to_snake

try:
    from jinja2 import Environment

    from sanic_ext.extensions.templating.engine import Templating
    from sanic_ext.extensions.templating.extension import TemplatingExtension

    TEMPLATING_ENABLED = True
except (ImportError, ModuleNotFoundError):
    TEMPLATING_ENABLED = False

MIN_SUPPORT = (21, 3, 2)


class Extend:
    _pre_registry: List[Union[Type[Extension], Extension]] = []

    if TEMPLATING_ENABLED:
        environment: Environment
        templating: Templating

    def __init__(
        self,
        app: Sanic,
        *,
        extensions: Optional[List[Union[Type[Extension], Extension]]] = None,
        built_in_extensions: bool = True,
        config: Optional[Union[Config, Dict[str, Any]]] = None,
        **kwargs,
    ) -> None:
        """
        Ingress for instantiating sanic-ext

        :param app: Sanic application
        :type app: Sanic
        """
        if not isinstance(app, Sanic):
            raise SanicException(
                f"Cannot apply SanicExt to {app.__class__.__name__}"
            )

        sanic_version = tuple(
            map(int, __version__.strip(ascii_lowercase).split(".", 3)[:3])
        )

        if MIN_SUPPORT > sanic_version:
            min_version = ".".join(map(str, MIN_SUPPORT))
            raise SanicException(
                f"Sanic Extensions only works with Sanic v{min_version} "
                f"and above. It looks like you are running {__version__}."
            )

        self._injection_registry: Optional[InjectionRegistry] = None
        self._openapi: Optional[SpecificationBuilder] = None
        self.app = app
        self.extensions: List[Extension] = []
        self.sanic_version = sanic_version
        app._ext = self
        app.ctx._dependencies = SimpleNamespace()

        if not isinstance(config, Config):
            config = Config.from_dict(config or {})
        self.config = add_fallback_config(app, config, **kwargs)

        extensions = extensions or []
        extensions.extend(Extend._pre_registry)
        if built_in_extensions:
            extensions.extend(
                [
                    InjectionExtension,
                    OpenAPIExtension,
                    HTTPExtension,
                    HealthExtension,
                    LoggingExtension,
                ]
            )

            if TEMPLATING_ENABLED:
                extensions.append(TemplatingExtension)

        started = set()
        for ext in extensions:
            if ext in started:
                continue
            extension = Extension.create(ext, app, self.config)
            extension._startup(self)
            self.extensions.append(extension)
            started.add(ext)

    def _display(self):
        init_logs = ["Sanic Extensions:"]
        for extension in self.extensions:
            label = extension.render_label()
            if extension.included():
                init_logs.append(f"  > {extension.name} {label}")

        list(map(logger.info, init_logs))

    def injection(
        self,
        type: Type,
        constructor: Optional[Callable[..., Any]] = None,
    ) -> None:
        warn(
            "The 'ext.injection' method has been deprecated and will be "
            "removed in v22.6. Please use 'ext.add_dependency' instead.",
            DeprecationWarning,
        )
        self.add_dependency(type=type, constructor=constructor)

    def add_dependency(
        self,
        type: Type,
        constructor: Optional[Callable[..., Any]] = None,
    ) -> None:
        if not self._injection_registry:
            raise SanicException("Injection extension not enabled")
        self._injection_registry.register(type, constructor)

    def dependency(self, obj: Any, name: Optional[str] = None) -> None:
        if not name:
            name = camel_to_snake(obj.__class__.__name__)
        setattr(self.app.ctx._dependencies, name, obj)

        def getter(*_):
            return obj

        self.add_dependency(obj.__class__, getter)

    @property
    def openapi(self) -> SpecificationBuilder:
        if not self._openapi:
            self._openapi = SpecificationBuilder()

        return self._openapi

    def template(self, template_name: str, **kwargs):
        return self.templating.template(template_name, **kwargs)

    @classmethod
    def register(cls, extension: Union[Type[Extension], Extension]) -> None:
        cls._pre_registry.append(extension)

    @classmethod
    def reset(cls) -> None:
        cls._pre_registry.clear()



from dataclasses import dataclass

from sanic import Request, Sanic
from sanic.response import text

from sanic_ext import openapi
from sanic_ext.extensions.openapi.definitions import RequestBody

from .utils import get_spec


@dataclass
class UserProfile:
    name: str
    age: int
    email: str


def test_body(app: Sanic):
    @app.route("/test0")
    async def handler0(request: Request):
        """
        openapi:
        ---
        summary: Updates a pet in the store with form data
        description: some description
        operationId: someId
        requestBody:
          description: user to add to the system
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - name
                  - age
                  - email
                properties:
                  name:
                    type: string
                  email:
                    type: string
                  age:
                    type: integer
                    format: int32
                    minimum: 0
        """
        return text("ok")

    @app.route("/test1")
    @openapi.body({"application/json": UserProfile})
    async def handler1(request: Request):
        return text("ok")

    @app.route("/test2")
    @openapi.body(RequestBody({"application/json": UserProfile}))
    async def handler2(request: Request):
        return text("ok")

    @app.route("/test3")
    @openapi.body(RequestBody(UserProfile))
    async def handler3(request: Request):
        return text("ok")

    @app.route("/test4")
    @openapi.body(UserProfile)
    async def handler4(request: Request):
        return text("ok")

    spec = get_spec(app)
    for i in range(5):
        assert f"/test{i}" in spec["paths"]
        content = spec["paths"][f"/test{i}"]["get"]["requestBody"]["content"]
        media_type = next(iter(content.keys()))
        if i <= 2:
            assert media_type == "application/json"
        else:
            assert media_type == "*/*"
        properties = content[media_type]["schema"]["properties"]
        assert properties["name"]["type"] == "string"
        assert properties["age"]["type"] == "integer"
        assert properties["email"]["type"] == "string"



from typing import Type, Union
from unittest.mock import Mock

import pytest
from sanic import Sanic

from sanic_ext import Extend, Extension


def test_multiple_extensions(bare_app: Sanic):
    mock = Mock()

    class FooExtension(Extension):
        name = "foo"

        def __init__(self, *args, **kwargs) -> None:
            super().__init__(*args, **kwargs)

        def startup(self, _) -> None:
            mock()

    class BarExtension(Extension):
        name = "bar"

        def __init__(self, *args, **kwargs) -> None:
            super().__init__(*args, **kwargs)

        def startup(self, _) -> None:
            mock()

    bare_app.extend(extensions=[FooExtension, BarExtension, BarExtension])

    assert mock.call_count == 2


@pytest.mark.parametrize("instance", (True, False))
def test_multiple_extensions_pre_register(bare_app: Sanic, instance: bool):
    mock = Mock()

    class FooExtension(Extension):
        name = "foo"

        def startup(self, _) -> None:
            mock()

    class BarExtension(Extension):
        name = "bar"

        def startup(self, _) -> None:
            mock()

    foo: Union[Type[Extension], Extension] = FooExtension
    bar: Union[Type[Extension], Extension] = BarExtension
    if instance:
        foo = foo()  # type: ignore
        bar = bar()  # type: ignore

    Extend.register(foo)
    Extend.register(bar)
    bare_app.extend()

    assert mock.call_count == 2



import sys
from dataclasses import dataclass
from typing import List, Optional

import pytest
from sanic import json
from sanic.views import HTTPMethodView

from sanic_ext import validate
from sanic_ext.extras.validation.check import check_data
from sanic_ext.extras.validation.schema import make_schema, parse_hint

from . import __models__ as models


def test_schema():
    @dataclass
    class Pet:
        name: str

    @dataclass
    class Person:
        name: str
        age: int
        pets: Optional[List[Pet]]

    schema = make_schema({}, Person)

    assert "Person" in schema
    assert schema["Person"]["hints"]["name"] == parse_hint(str)
    assert schema["Person"]["hints"]["age"] == parse_hint(int)
    assert schema["Person"]["hints"]["pets"] == parse_hint(Optional[List[Pet]])

    assert "Pet" in schema
    assert schema["Pet"]["hints"]["name"] == parse_hint(str)


def test_should_hydrate():
    @dataclass
    class Pet:
        name: str

    @dataclass
    class Person:
        name: str
        age: int
        pets: List[Pet]

    data = {"name": "Charlie Brown", "age": 8, "pets": [{"name": "Snoopy"}]}

    schema = make_schema({}, Person)
    cb = check_data(Person, data, schema)

    assert cb.name == "Charlie Brown"
    assert cb.age == 8
    assert cb.pets[0].name == "Snoopy"


@pytest.mark.parametrize(
    "data",
    (
        {"name": "Charlie Brown", "age": 8, "pets": {"name": "Snoopy"}},
        {"name": "Charlie Brown", "age": 8, "pets": [{"name": 123}]},
        {"name": "Charlie Brown", "age": 8, "pets": [123]},
        {"name": "Charlie Brown", "age": 8, "pets": 123},
        {"name": "Charlie Brown", "age": "8", "pets": {"name": "Snoopy"}},
        {"name": True, "age": 8, "pets": {"name": "Snoopy"}},
    ),
)
def test_should_not_hydrate(data):
    @dataclass
    class Pet:
        name: str

    @dataclass
    class Person:
        name: str
        age: int
        pets: List[Pet]

    schema = make_schema({}, Person)
    with pytest.raises(TypeError):
        check_data(Person, data, schema)


@pytest.mark.parametrize(
    "model,okay,data",
    (
        (models.ModelStr, True, {"foo": "bar"}),
        (models.ModelStr, False, {"foo": 1}),
        (models.ModelStr, False, {"foo": True}),
        (models.ModelStr, False, {"foo": ["bar"]}),
        (models.ModelStr, False, {"bar": "bar"}),
        (models.ModelStr, False, {"foo": None}),
        (models.ModelStr, False, 123),
        (models.ModelInt, True, {"foo": 1}),
        (models.ModelInt, True, {"foo": True}),
        (models.ModelInt, False, {"foo": "1"}),
        (models.ModelInt, False, {"foo": 1.1}),
        (models.ModelInt, False, {"foo": None}),
        (models.ModelFloat, True, {"foo": 1.1}),
        (models.ModelFloat, False, {"foo": 1}),
        (models.ModelFloat, False, {"foo": "1.1"}),
        (models.ModelFloat, False, {"foo": None}),
        (models.ModelBool, True, {"foo": True}),
        (models.ModelBool, True, {"foo": False}),
        (models.ModelBool, False, {"foo": 1}),
        (models.ModelBool, False, {"foo": 0}),
        (models.ModelBool, False, {"foo": 2}),
        (models.ModelBool, False, {"foo": "True"}),
        (models.ModelBool, False, {"foo": None}),
        (models.ModelOptionalStr, True, {"foo": "bar"}),
        (models.ModelOptionalStr, True, {"foo": None}),
        (models.ModelOptionalStr, False, {"foo": 0}),
        (models.ModelUnion, True, {"foo": 1}),
        (models.ModelUnion, True, {"foo": 1.1}),
        (models.ModelUnion, False, {"foo": "1.1"}),
        (models.ModelUnion, False, {"foo": None}),
        (models.ModelUnionModels, True, {"foo": {"foo": 1}}),
        (models.ModelUnionModels, True, {"foo": {"foo": 1.1}}),
        (models.ModelUnionModels, False, {"foo": {"foo": "1.1"}}),
        (models.ModelUnionModels, False, {"foo": 1}),
        (models.ModelUnionModels, False, {"foo": 1.1}),
        (models.ModelUnionModels, False, {"foo": None}),
        (models.ModelUnionStrInt, True, {"foo": "1"}),
        (models.ModelUnionStrInt, True, {"foo": "1q"}),
        (models.ModelUnionStrInt, True, {"foo": 1}),
        (models.ModelUnionStrInt, False, {"foo": 1.1}),
        (models.ModelUnionStrInt, False, {"foo": None}),
        (models.ModelUnionIntStr, True, {"foo": "1"}),
        (models.ModelUnionIntStr, True, {"foo": "1q"}),
        (models.ModelUnionIntStr, True, {"foo": 1}),
        (models.ModelUnionIntStr, False, {"foo": 1.1}),
        (models.ModelUnionIntStr, False, {"foo": None}),
        (models.ModelOptionalUnionStrInt, True, {"foo": "1"}),
        (models.ModelOptionalUnionStrInt, True, {"foo": "1q"}),
        (models.ModelOptionalUnionStrInt, True, {"foo": 1}),
        (models.ModelOptionalUnionStrInt, False, {"foo": 1.1}),
        (models.ModelOptionalUnionStrInt, True, {"foo": None}),
        (models.ModelOptionalUnionIntStr, True, {"foo": "1"}),
        (models.ModelOptionalUnionIntStr, True, {"foo": "1q"}),
        (models.ModelOptionalUnionIntStr, True, {"foo": 1}),
        (models.ModelOptionalUnionIntStr, False, {"foo": 1.1}),
        (models.ModelOptionalUnionIntStr, True, {"foo": None}),
        (models.ModelListStr, True, {"foo": ["bar"]}),
        (models.ModelListStr, True, {"foo": ["one", "two"]}),
        (models.ModelListStr, False, {"foo": "bar"}),
        (models.ModelListStr, False, {"foo": ["one", 2]}),
        (models.ModelListStr, False, {"foo": ["one", None]}),
        (models.ModelListStr, False, {"foo": None}),
        (models.ModelListModel, True, {"foo": [{"foo": "bar"}]}),
        (
            models.ModelListModel,
            True,
            {"foo": [{"foo": "one"}, {"foo": "two"}]},
        ),
        (models.ModelListModel, False, {"foo": {"foo": "bar"}}),
        (models.ModelListModel, False, {"foo": [{"foo": "bar"}, 2]}),
        (models.ModelListModel, False, {"foo": [{"foo": "bar"}, None]}),
        (models.ModelListModel, False, {"foo": None}),
        (models.ModelOptionalList, True, {"foo": None}),
        (models.ModelOptionalList, True, {"foo": ["bar"]}),
        (models.ModelOptionalList, False, {"foo": [1]}),
        (models.ModelOptionalList, False, {"foo": [None]}),
        (models.ModelListUnion, True, {"foo": [1]}),
        (models.ModelListUnion, True, {"foo": [1.1]}),
        (models.ModelListUnion, True, {"foo": [1, 1.1]}),
        (models.ModelListUnion, False, {"foo": [1, 1.1, "one"]}),
        (models.ModelListUnion, False, {"foo": [1, 1.1, None]}),
        (models.ModelListUnion, False, {"foo": 1}),
        (models.ModelListUnion, False, {"foo": 1.1}),
        (models.ModelListUnion, False, {"foo": None}),
        (models.ModelOptionalListUnion, True, {"foo": [1]}),
        (models.ModelOptionalListUnion, True, {"foo": [1.1]}),
        (models.ModelOptionalListUnion, True, {"foo": [1, 1.1]}),
        (models.ModelOptionalListUnion, True, {"foo": None}),
        (models.ModelOptionalListUnion, False, {"foo": [1, 1.1, "one"]}),
        (models.ModelOptionalListUnion, False, {"foo": [1, 1.1, None]}),
        (models.ModelOptionalListUnion, False, {"foo": 1}),
        (models.ModelOptionalListUnion, False, {"foo": 1.1}),
        (models.ModelModel, True, {"foo": {"foo": "one"}}),
        (models.ModelModel, False, {"foo": {"foo": 1}}),
        (models.ModelModel, False, {"foo": {"foo": None}}),
        (models.ModelModel, False, {"foo": "one"}),
        (models.ModelModel, False, {"foo": None}),
        (models.ModelOptionalModel, True, {"foo": {"foo": "one"}}),
        (models.ModelOptionalModel, True, {"foo": None}),
        (models.ModelOptionalModel, False, {"foo": {"foo": 1}}),
        (models.ModelOptionalModel, False, {"foo": {"foo": None}}),
        (models.ModelOptionalModel, False, {"foo": "one"}),
        (models.ModelDictStr, True, {"foo": {"foo": "one"}}),
        (models.ModelDictStr, False, {"foo": {"foo": 1}}),
        (models.ModelDictStr, False, {"foo": {"foo": None}}),
        (models.ModelDictStr, False, {"foo": "one"}),
        (models.ModelDictStr, False, {"foo": None}),
        (models.ModelDictModel, True, {"foo": {"foo": {"foo": "one"}}}),
        (models.ModelDictModel, False, {"foo": {"foo": {"foo": 1}}}),
        (models.ModelDictModel, False, {"foo": {"foo": 1}}),
        (models.ModelDictModel, False, {"foo": {"foo": None}}),
        (models.ModelDictModel, False, {"foo": "one"}),
        (models.ModelDictModel, False, {"foo": None}),
        (models.ModelOptionalDict, True, {"foo": {"foo": "one"}}),
        (models.ModelOptionalDict, True, {"foo": None}),
        (models.ModelOptionalDict, False, {"foo": {"foo": 1}}),
        (models.ModelOptionalDict, False, {"foo": {"foo": None}}),
        (models.ModelOptionalDict, False, {"foo": "one"}),
        (models.ModelDictUnion, True, {"foo": {"foo": 1}}),
        (models.ModelDictUnion, True, {"foo": {"foo": 1.1}}),
        (models.ModelDictUnion, False, {"foo": {"foo": "one"}}),
        (models.ModelDictUnion, False, {"foo": {"foo": None}}),
        (models.ModelDictUnion, False, {"foo": "one"}),
        (models.ModelDictUnion, False, {"foo": 1}),
        (models.ModelDictUnion, False, {"foo": 1.1}),
        (models.ModelDictUnion, False, {"foo": None}),
        (models.ModelOptionalDictUnion, True, {"foo": {"foo": 1}}),
        (models.ModelOptionalDictUnion, True, {"foo": {"foo": 1.1}}),
        (models.ModelOptionalDictUnion, True, {"foo": None}),
        (models.ModelOptionalDictUnion, False, {"foo": {"foo": "one"}}),
        (models.ModelOptionalDictUnion, False, {"foo": {"foo": None}}),
        (models.ModelOptionalDictUnion, False, {"foo": "one"}),
        (models.ModelOptionalDictUnion, False, {"foo": 1}),
        (models.ModelOptionalDictUnion, False, {"foo": 1.1}),
        (models.ModelSingleLiteral, True, {"foo": True}),
        (models.ModelSingleLiteral, False, {"foo": False}),
        (models.ModelSingleLiteral, False, {"foo": "True"}),
        (models.ModelSingleLiteral, False, {"foo": None}),
        (models.ModelOptionalSingleLiteral, True, {"foo": True}),
        (models.ModelOptionalSingleLiteral, True, {"foo": None}),
        (models.ModelOptionalSingleLiteral, False, {"foo": False}),
        (models.ModelOptionalSingleLiteral, False, {"foo": "True"}),
        (models.ModelOptionalMultipleLiteral, True, {"foo": True}),
        (models.ModelOptionalMultipleLiteral, True, {"foo": 1}),
        (models.ModelOptionalMultipleLiteral, True, {"foo": "y"}),
        (models.ModelOptionalMultipleLiteral, True, {"foo": "Y"}),
        (models.ModelOptionalMultipleLiteral, True, {"foo": None}),
        (models.ModelOptionalMultipleLiteral, False, {"foo": "n"}),
        (models.ModelOptionalMultipleLiteral, False, {"foo": False}),
        (models.ModelListStrWithDefaultFactory, True, {}),
        (models.ModelListStrWithDefaultFactory, True, {"foo": ["bar"]}),
        (models.ModelListStrWithDefaultFactory, True, {"foo": []}),
        (models.ModelListStrWithDefaultFactory, False, {"foo": [1]}),
        (models.ModelListStrWithDefaultFactory, False, {"foo": None}),
    ),
)
def test_modeling(model, okay, data):
    schema = make_schema({}, model)

    if okay:
        check_data(model, data, schema)
    else:
        with pytest.raises(TypeError):
            check_data(model, data, schema)


@pytest.mark.skipif(
    sys.version_info < (3, 10), reason="UnionType added in 3.10"
)
def test_modeling_union_type_ModelUnionTypeStrNone():
    schema = make_schema({}, models.ModelUnionTypeStrNone)

    check_data(models.ModelUnionTypeStrNone, {"foo": "bar"}, schema)
    check_data(models.ModelUnionTypeStrNone, {"foo": None}, schema)
    with pytest.raises(TypeError):
        check_data(models.ModelUnionTypeStrNone, {"foo": 0}, schema)


@pytest.mark.skipif(
    sys.version_info < (3, 10), reason="UnionType added in 3.10"
)
def test_modeling_union_type_ModelUnionTypeStrIntNone():
    schema = make_schema({}, models.ModelUnionTypeStrIntNone)

    check_data(models.ModelUnionTypeStrIntNone, {"foo": "1"}, schema)
    check_data(models.ModelUnionTypeStrIntNone, {"foo": "bar"}, schema)
    check_data(models.ModelUnionTypeStrIntNone, {"foo": None}, schema)
    check_data(models.ModelUnionTypeStrIntNone, {"foo": 1}, schema)
    check_data(models.ModelUnionTypeStrIntNone, {"foo": 0}, schema)
    with pytest.raises(TypeError):
        check_data(models.ModelUnionTypeStrIntNone, {"foo": 1.1}, schema)


@pytest.mark.skipif(
    sys.version_info < (3, 10), reason="UnionType added in 3.10"
)
def test_modeling_union_type_ModelUnionTypeStrInt():
    schema = make_schema({}, models.ModelUnionTypeStrInt)

    check_data(models.ModelUnionTypeStrInt, {"foo": "1"}, schema)
    check_data(models.ModelUnionTypeStrInt, {"foo": "bar"}, schema)
    check_data(models.ModelUnionTypeStrInt, {"foo": 1}, schema)
    check_data(models.ModelUnionTypeStrInt, {"foo": 0}, schema)
    with pytest.raises(TypeError):
        check_data(models.ModelUnionTypeStrInt, {"foo": None}, schema)
    with pytest.raises(TypeError):
        check_data(models.ModelUnionTypeStrInt, {"foo": 1.1}, schema)


def test_validate_decorator(app):
    @dataclass
    class Pet:
        name: str

    @app.post("/function")
    @validate(json=Pet)
    async def handler(_, body: Pet):
        return json({"is_pet": isinstance(body, Pet)})

    class MethodView(HTTPMethodView, attach=app, uri="/method"):
        decorators = [validate(json=Pet)]

        async def post(self, _, body: Pet):
            return json({"is_pet": isinstance(body, Pet)})

    _, response = app.test_client.post("/function", json={"name": "Snoopy"})
    assert response.status == 200
    assert response.json["is_pet"]

    _, response = app.test_client.post("/method", json={"name": "Snoopy"})
    assert response.status == 200
    assert response.json["is_pet"]



import json
import uuid
from datetime import date, datetime, time
from enum import Enum
from inspect import getmembers, isclass, isfunction, ismethod
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

from sanic_routing.patterns import nonemptystr

from sanic_ext.utils.typing import is_generic


class Definition:
    __nullable__: Optional[List[str]] = []
    __ignore__: Optional[List[str]] = []

    def __init__(self, **kwargs):
        self._fields: Dict[str, Any] = self.guard(kwargs)

    @property
    def fields(self):
        return self._fields

    def guard(self, fields):
        return {
            k: v
            for k, v in fields.items()
            if k in _properties(self).keys() or k.startswith("x-")
        }

    def serialize(self):
        return {
            k: self._value(v)
            for k, v in _serialize(self.fields).items()
            if (
                k not in self.__ignore__
                and (
                    v is not None
                    or (
                        isinstance(self.__nullable__, list)
                        and (not self.__nullable__ or k in self.__nullable__)
                    )
                )
            )
        }

    def __str__(self):
        return json.dumps(self.serialize())

    @staticmethod
    def _value(value):
        if isinstance(value, Enum):
            return value.value
        return value


class Schema(Definition):
    title: str
    description: str
    type: str
    format: str
    nullable: bool
    required: bool
    default: None
    example: None
    oneOf: List[Definition]
    anyOf: List[Definition]
    allOf: List[Definition]

    additionalProperties: Dict[str, str]
    multipleOf: int
    maximum: int
    exclusiveMaximum: bool
    minimum: int
    exclusiveMinimum: bool
    maxLength: int
    minLength: int
    pattern: str
    enum: Union[List[Any], Enum]

    @staticmethod
    def make(value, **kwargs):
        _type = type(value)
        origin = get_origin(value)
        args = get_args(value)
        if origin is Union:
            if type(None) in args:
                kwargs["nullable"] = True

            filtered = [arg for arg in args if arg is not type(None)]  # noqa

            if len(filtered) == 1:
                return Schema.make(filtered[0], **kwargs)
            return Schema(
                oneOf=[Schema.make(arg) for arg in filtered], **kwargs
            )
            # return Schema.make(value, **kwargs)

        if isinstance(value, Schema):
            return value
        if value == bool:
            return Boolean(**kwargs)
        elif value == int:
            return Integer(**kwargs)
        elif value == float:
            return Float(**kwargs)
        elif value == str or value is nonemptystr:
            return String(**kwargs)
        elif value == bytes:
            return Byte(**kwargs)
        elif value == bytearray:
            return Binary(**kwargs)
        elif value == date:
            return Date(**kwargs)
        elif value == time:
            return Time(**kwargs)
        elif value == datetime:
            return DateTime(**kwargs)
        elif value == uuid.UUID:
            return UUID(**kwargs)
        elif value == Any:
            return AnyValue(**kwargs)

        if _type == bool:
            return Boolean(default=value, **kwargs)
        elif _type == int:
            return Integer(default=value, **kwargs)
        elif _type == float:
            return Float(default=value, **kwargs)
        elif _type == str:
            return String(default=value, **kwargs)
        elif _type == bytes:
            return Byte(default=value, **kwargs)
        elif _type == bytearray:
            return Binary(default=value, **kwargs)
        elif _type == date:
            return Date(**kwargs)
        elif _type == time:
            return Time(**kwargs)
        elif _type == datetime:
            return DateTime(**kwargs)
        elif _type == uuid.UUID:
            return UUID(**kwargs)
        elif _type == list:
            if len(value) == 0:
                schema = Schema(nullable=True)
            elif len(value) == 1:
                schema = Schema.make(value[0])
            else:
                schema = Schema(oneOf=[Schema.make(x) for x in value])

            return Array(schema, **kwargs)
        elif _type == dict:
            return Object.make(value, **kwargs)
        elif (
            (is_generic(value) or is_generic(_type))
            and origin == dict
            and len(args) == 2
        ):
            kwargs["additionalProperties"] = Schema.make(args[1])
            return Object(**kwargs)
        elif (is_generic(value) or is_generic(_type)) and origin == list:
            return Array(Schema.make(args[0]), **kwargs)
        elif _type is type(Enum):
            available = [item.value for item in value.__members__.values()]
            available_types = list({type(item) for item in available})
            schema_type = (
                available_types[0] if len(available_types) == 1 else "string"
            )
            return Schema.make(
                schema_type,
                enum=[item.value for item in value.__members__.values()],
            )
        else:
            return Object.make(value, **kwargs)


class Boolean(Schema):
    def __init__(self, **kwargs):
        super().__init__(type="boolean", **kwargs)


class Integer(Schema):
    def __init__(self, **kwargs):
        super().__init__(type="integer", format="int32", **kwargs)


class Long(Schema):
    def __init__(self, **kwargs):
        super().__init__(type="integer", format="int64", **kwargs)


class Float(Schema):
    def __init__(self, **kwargs):
        super().__init__(type="number", format="float", **kwargs)


class Double(Schema):
    def __init__(self, **kwargs):
        super().__init__(type="number", format="double", **kwargs)


class String(Schema):
    def __init__(self, **kwargs):
        super().__init__(type="string", **kwargs)


class Byte(Schema):
    def __init__(self, **kwargs):
        super().__init__(type="string", format="byte", **kwargs)


class Binary(Schema):
    def __init__(self, **kwargs):
        super().__init__(type="string", format="binary", **kwargs)


class Date(Schema):
    def __init__(self, **kwargs):
        super().__init__(type="string", format="date", **kwargs)


class Time(Schema):
    def __init__(self, **kwargs):
        super().__init__(type="string", format="time", **kwargs)


class DateTime(Schema):
    def __init__(self, **kwargs):
        super().__init__(type="string", format="date-time", **kwargs)


class Password(Schema):
    def __init__(self, **kwargs):
        super().__init__(type="string", format="password", **kwargs)


class Email(Schema):
    def __init__(self, **kwargs):
        super().__init__(type="string", format="email", **kwargs)


class UUID(Schema):
    def __init__(self, **kwargs):
        super().__init__(type="string", format="uuid", **kwargs)


class AnyValue(Schema):
    @classmethod
    def make(cls, value: Any, **kwargs):
        return cls(
            AnyValue={},
            **kwargs,
        )


class Object(Schema):
    properties: Dict[str, Schema]
    maxProperties: int
    minProperties: int

    def __init__(
        self, properties: Optional[Dict[str, Schema]] = None, **kwargs
    ):
        if properties:
            kwargs["properties"] = properties
        super().__init__(type="object", **kwargs)

    @classmethod
    def make(cls, value: Any, **kwargs):
        return cls(
            {k: Schema.make(v) for k, v in _properties(value).items()},
            **kwargs,
        )


class Array(Schema):
    items: Any
    maxItems: int
    minItems: int
    uniqueItems: bool

    def __init__(self, items: Any, **kwargs):
        super().__init__(type="array", items=Schema.make(items), **kwargs)


def _serialize(value) -> Any:
    if isinstance(value, Definition):
        return value.serialize()

    if isinstance(value, type) and issubclass(value, Enum):
        return [item.value for item in value.__members__.values()]

    if isinstance(value, dict):
        return {k: _serialize(v) for k, v in value.items()}

    if isinstance(value, list):
        return [_serialize(v) for v in value]

    return value


def _properties(value: object) -> Dict:
    try:
        fields = {
            x: val
            for x, v in getmembers(value, _is_property)
            if (val := _extract(v)) and x in value.__dict__
        }
    except AttributeError:
        fields = {}

    cls = value if callable(value) else value.__class__
    return {
        k: v
        for k, v in {**get_type_hints(cls), **fields}.items()
        if not k.startswith("_")
    }


def _extract(item):
    if isinstance(item, property):
        hints = get_type_hints(item.fget)
        return hints.get("return")
    return item


def _is_property(item):
    return not isfunction(item) and not ismethod(item) and not isclass(item)



"""
Classes defined from the OpenAPI 3.0 specifications.

I.e., the objects described https://swagger.io/docs/specification

"""
from __future__ import annotations

from inspect import isclass
from typing import (
    Any,
    Dict,
    List,
    Literal,
    Optional,
    Type,
    Union,
    get_type_hints,
)

from sanic.exceptions import SanicException

from .types import Definition, Schema


class Reference(Schema):
    def __init__(self, value):
        super().__init__(**{"$ref": value})

    def guard(self, fields: Dict[str, Any]):
        return fields


class Contact(Definition):
    name: str
    url: str
    email: str


class License(Definition):
    name: str
    url: str

    def __init__(self, name: str, **kwargs):
        super().__init__(name=name, **kwargs)


class Info(Definition):
    title: str
    description: str
    termsOfService: str
    contact: Contact
    license: License
    version: str

    def __init__(self, title: str, version: str, **kwargs):
        super().__init__(title=title, version=version, **kwargs)


class Example(Definition):
    summary: str
    description: str
    value: Any
    externalValue: str

    def __init__(self, value: Any = None, **kwargs):
        super().__init__(value=value, **kwargs)

    @staticmethod
    def make(value: Any, **kwargs):
        return Example(Schema.make(value), **kwargs)

    @staticmethod
    def external(value: Any, **kwargs):
        return Example(externalValue=value, **kwargs)


class MediaType(Definition):
    schema: Schema
    example: Any

    def __init__(self, schema: Schema, **kwargs):
        super().__init__(schema=schema, **kwargs)

    @staticmethod
    def make(value: Any):
        if isinstance(value, dict):
            kwargs = {}
            if "schema" in value:
                kwargs = {**value}
                value = kwargs.pop("schema")
            return MediaType(value, **kwargs)
        return MediaType(Schema.make(value))

    @staticmethod
    def all(content: Any):
        media_types = (
            content if isinstance(content, dict) else {"*/*": content or {}}
        )

        return {x: MediaType.make(v) for x, v in media_types.items()}


class Response(Definition):
    content: Union[Any, Dict[str, Union[Any, MediaType]]]
    description: Optional[str]
    status: Union[Literal["default"], int]

    __nullable__ = None
    __ignore__ = ["status"]

    def __init__(
        self,
        content: Optional[Union[Any, Dict[str, Union[Any, MediaType]]]] = None,
        status: Union[Literal["default"], int] = "default",
        description: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(
            content=content,
            status=status,
            description=description,
            **kwargs,
        )

    @staticmethod
    def make(content, description: str = None, **kwargs):
        if not description:
            description = "Default Response"

        return Response(
            MediaType.all(content), description=description, **kwargs
        )


class RequestBody(Definition):
    description: Optional[str]
    required: Optional[bool]
    content: Union[Any, Dict[str, Union[Any, MediaType]]]

    __nullable__ = None

    def __init__(
        self,
        content: Union[Any, Dict[str, Union[Any, MediaType]]],
        required: Optional[bool] = None,
        description: Optional[str] = None,
        **kwargs,
    ):
        """Can be initialized with content in one of a few ways:

        RequestBody(SomeModel)
        RequestBody({"application/json": SomeModel})
        RequestBody({"application/json": {"name": str}})
        """
        super().__init__(
            content=content,
            required=required,
            description=description,
            **kwargs,
        )

    @staticmethod
    def make(content: Any, **kwargs):
        return RequestBody(MediaType.all(content), **kwargs)


class ExternalDocumentation(Definition):
    url: str
    description: str

    __nullable__ = None

    def __init__(self, url: str, description=None):
        super().__init__(url=url, description=description)

    @staticmethod
    def make(url: str, description: str = None):
        return ExternalDocumentation(url, description)


class Header(Definition):
    name: str
    description: str
    externalDocs: ExternalDocumentation

    def __init__(self, url: str, description=None):
        super().__init__(url=url, description=description)

    @staticmethod
    def make(url: str, description: str = None):
        return Header(url, description)


class Parameter(Definition):
    name: str
    schema: Union[Type, Schema]
    location: str
    description: Optional[str]
    required: Optional[bool]
    deprecated: Optional[bool]
    allowEmptyValue: Optional[bool]

    __nullable__ = None

    def __init__(
        self,
        name: str,
        schema: Union[Type, Schema] = str,
        location: str = "query",
        description: Optional[str] = None,
        required: Optional[bool] = None,
        deprecated: Optional[bool] = None,
        allowEmptyValue: Optional[bool] = None,
        **kwargs,
    ):
        super().__init__(
            name=name,
            schema=schema,
            location=location,
            description=description,
            required=required,
            deprecated=deprecated,
            allowEmptyValue=allowEmptyValue,
            **kwargs,
        )

    @property
    def fields(self):
        values = super().fields

        if "location" in values:
            values["in"] = values.pop("location")

        return values

    @staticmethod
    def make(name: str, schema: type, location: str, **kwargs):
        if location == "path" and "required" not in kwargs:
            kwargs["required"] = True

        return Parameter(name, Schema.make(schema), location, **kwargs)


class Operation(Definition):
    tags: List[str]
    summary: str
    description: str
    operationId: str
    requestBody: RequestBody
    externalDocs: ExternalDocumentation
    parameters: List[Parameter]
    responses: Dict[str, Response]
    security: Dict[str, List[str]]
    callbacks: List[str]  # TODO
    deprecated: bool
    servers: List[Dict[str, str]]


class PathItem(Definition):
    summary: str
    description: str
    get: Operation
    put: Operation
    post: Operation
    delete: Operation
    options: Operation
    head: Operation
    patch: Operation
    trace: Operation


class Flow(Definition):
    authorizationUrl: str
    tokenUrl: str
    refreshUrl: str
    scopes: Dict[str, str]


class Flows(Definition):
    implicit: Flow
    password: Flow
    clientCredentials: Flow
    authorizationCode: Flow


class SecurityRequirement(Definition):
    name: str
    value: List[str]


class SecurityScheme(Definition):
    type: str
    bearerFormat: str
    description: str
    flows: Flows
    location: str
    name: str
    openIdConnectUrl: str
    scheme: str

    __nullable__ = None

    def __init__(self, type: str, **kwargs):
        super().__init__(type=type, **kwargs)

    @property
    def fields(self):
        values = super().fields

        if "location" in values:
            values["in"] = values.pop("location")

        return values

    @staticmethod
    def make(_type: str, cls: Type, **kwargs):
        params = cls.__dict__ if hasattr(cls, "__dict__") else {}

        return SecurityScheme(_type, **params, **kwargs)


class ServerVariable(Definition):
    default: str
    description: str
    enum: List[str]

    def __init__(self, default: str, **kwargs):
        super().__init__(default=default, **kwargs)


class Server(Definition):
    url: str
    description: str
    variables: Dict[str, ServerVariable]

    __nullable__ = None

    def __init__(
        self, url: str, description: str = None, variables: dict = None
    ):
        super().__init__(
            url=url, description=description, variables=variables or []
        )


class Tag(Definition):
    name: str
    description: str
    externalDocs: ExternalDocumentation

    def __init__(self, name: str, **kwargs):
        super().__init__(name=name, **kwargs)


class Components(Definition):
    # This class is not being used in sanic-openapi right now, but the
    # definition is kept here to keep in close accordance with the openapi
    # spec, in case it is desired to be added later.
    schemas: Dict[str, Schema]
    responses: Dict[str, Response]
    parameters: Dict[str, Parameter]
    examples: Dict[str, Example]
    requestBodies: Dict[str, RequestBody]
    headers: Dict[str, Header]
    securitySchemes: Dict[str, SecurityScheme]
    links: Dict[str, Schema]  # TODO
    callbacks: Dict[str, Schema]  # TODO


def Component(
    obj: Any, *, name: str = "", field: str = "schemas"
) -> Reference:
    hints = get_type_hints(Components)

    if field not in hints:
        raise AttributeError(
            f"Unknown field '{field}'. Must be a valid field per OAS3 "
            "requirements. See "
            "https://swagger.io/specification/#components-object."
        )

    if not isclass(obj) and not name:
        raise SanicException(
            f"Components {obj} must be created with a declared name"
        )

    if not name:
        name = obj.__name__

    from sanic_ext.extensions.openapi.builders import SpecificationBuilder

    spec = SpecificationBuilder()
    refval = f"#/components/{field}/{name}"
    ref = Reference(refval)

    if not spec.has_component(field, name):
        prop_info = hints[field]
        type_ = prop_info.__args__[1]
        component = type_.make(obj) if hasattr(type_, "make") else type_(obj)

        spec.add_component(field, name, component)

    return ref


class OpenAPI(Definition):
    openapi: str
    info: Info
    servers: List[Server]
    paths: Dict[str, PathItem]
    components: Components
    security: Dict[str, SecurityScheme]
    tags: List[Tag]
    externalDocs: ExternalDocumentation

    def __init__(self, info: Info, paths: Dict[str, PathItem], **kwargs):
        use = {k: v for k, v in kwargs.items() if v is not None}
        super().__init__(openapi="3.0.3", info=info, paths=paths, **use)



from sanic.exceptions import SanicException

from ..base import Extension
from .logger import Logger


class LoggingExtension(Extension):
    name = "logging"
    MIN_VERSION = (22, 9)

    def startup(self, bootstrap) -> None:
        if self.included():
            if self.MIN_VERSION > bootstrap.sanic_version:
                min_version = ".".join(map(str, self.MIN_VERSION))
                sanic_version = ".".join(map(str, bootstrap.sanic_version))
                raise SanicException(
                    f"The logging extension only works with Sanic "
                    f"v{min_version} and above. It looks like you are "
                    f"running {sanic_version}."
                )
            Logger.setup(self.app)

    def included(self):
        return self.config.LOGGING



from sanic import Request, text


class Foo:
    def bar(self):
        return "foobar"


def test_dependency_added(app):
    foo = Foo()
    foobar = Foo()

    app.ext.dependency(foo)
    app.ext.dependency(foobar, name="something")

    assert app.ctx._dependencies.foo is foo
    assert app.ctx._dependencies.something is foobar


def test_dependency_injection(app):
    foo = Foo()

    app.ext.dependency(foo)

    @app.get("/getfoo")
    async def getfoo(request: Request, foo: Foo):
        return text(foo.bar())

    _, response = app.test_client.get("/getfoo")

    assert response.text == "foobar"



from sanic import Blueprint, Request, Sanic, text

from sanic_ext.extensions.openapi import openapi

from .utils import get_spec


def test_exclude_decorator(app: Sanic):
    @app.route("/test0")
    @openapi.exclude()
    async def handler0(request: Request):
        """
        openapi:
        ---
        summary: This is a summary.
        """
        return text("ok")

    @app.route("/test1")
    @openapi.definition(summary="This is a summary.", exclude=True)
    async def handler1(request: Request):
        return text("ok")

    spec = get_spec(app)
    paths = spec["paths"]
    assert len(paths) == 0


def test_exclude_bp(app: Sanic):
    bp1 = Blueprint("blueprint1")
    bp2 = Blueprint("blueprint2")

    @bp1.route("/op1")
    @openapi.summary("handler 1")
    async def handler1(request: Request):
        return text("bp1, ok")

    @bp2.route("/op2")
    @openapi.summary("handler 2")
    async def handler2(request: Request):
        return text("bp2, ok")

    app.blueprint(bp1)
    app.blueprint(bp2)

    openapi.exclude(bp=bp1)
    spec = get_spec(app)
    paths = spec["paths"]
    assert len(paths) == 1
    assert "/op2" in paths
    assert "/op1" not in paths
    assert paths["/op2"]["get"]["summary"] == "handler 2"



from pathlib import Path

from sanic import Sanic

from sanic_ext import render


def test_default_templates():
    app = Sanic("templating")
    app.extend(
        config={
            "templating_path_to_templates": Path(__file__).parent / "templates"
        }
    )

    @app.get("/1")
    @app.ext.template("foo.html")
    async def handler(_):
        return {"seq": ["one", "two"]}

    @app.get("/2")
    async def handler2(_):
        return await render(
            "foo.html", context={"seq": ["three", "four"]}, app=app
        )

    @app.get("/3")
    @app.ext.template("foo.html")
    async def handler3(_):
        return await render(
            context={"seq": ["five", "six"]}, status=201, app=app
        )

    _, response = app.test_client.get("/1")
    assert response.content_type == "text/html; charset=utf-8"
    assert "<li>one</li>" in response.text
    assert "<li>two</li>" in response.text

    _, response = app.test_client.get("/2")
    assert response.content_type == "text/html; charset=utf-8"
    assert "<li>three</li>" in response.text
    assert "<li>four</li>" in response.text

    _, response = app.test_client.get("/3")
    assert response.content_type == "text/html; charset=utf-8"
    assert "<li>five</li>" in response.text
    assert "<li>six</li>" in response.text
    assert response.status == 201


def test_render_from_string():
    app = Sanic("templating-from-string")
    app.extend()

    template = """
    <!DOCTYPE html>
<html lang="en">

    <head>
        <title>My Webpage</title>
    </head>

    <body>
        <h1>Hello, world!!!!</h1>
        <ul>
            {% for item in seq %}
            <li>{{ item }}</li>
            {% endfor %}
        </ul>
    </body>

</html>

    """

    @app.get("/2")
    async def handler2(_):
        return await render(
            template_source=template,
            context={"seq": ["three", "four"]},
            app=app,
        )

    _, response = app.test_client.get("/2")
    assert response.content_type == "text/html; charset=utf-8"
    assert "<li>three</li>" in response.text
    assert "<li>four</li>" in response.text


def test_config_templating_dir():
    app = Sanic("templating")
    app.config.TEMPLATING_PATH_TO_TEMPLATES = (
        Path(__file__).parent / "templates"
    )

    assert app.ext.templating.environment.get_template(
        "foo.html"
    ).filename == str(Path(__file__).parent / "templates" / "foo.html")



from unittest.mock import Mock

import pytest
from sanic.exceptions import SanicException
from sanic.signals import Event

from sanic_ext.config import Config
from sanic_ext.extensions.injection.injector import add_injection


def test_default_config_signal():
    config = Config()
    assert config.INJECTION_SIGNAL == Event.HTTP_ROUTING_AFTER


def test_not_allowed_signals_error():
    with pytest.raises(SanicException):
        Config(injection_signal=Event.HTTP_LIFECYCLE_REQUEST)


def test_http_routing_after_succeeds():
    Config(injection_signal=Event.HTTP_ROUTING_AFTER)


@pytest.mark.skipif(
    not hasattr(Event, "HTTP_HANDLER_BEFORE"),
    reason="Sanic version does not support HTTP_HANDLER_BEFORE",
)
def test_http_handler_before_succeeds():
    Config(injection_signal=Event.HTTP_HANDLER_BEFORE)


def test_add_injection_uses_signal_config():
    app = Mock()

    app.signal = Mock(return_value=Mock())
    app.ext.config.INJECTION_SIGNAL = "random_string"
    add_injection(app, Mock())

    app.signal.assert_called_once_with("random_string")



from collections import defaultdict
from logging import LogRecord
from logging.handlers import QueueHandler
from multiprocessing import Manager
from queue import Empty, Full
from signal import SIGINT, SIGTERM
from signal import signal as signal_func

from sanic import Sanic
from sanic.log import access_logger, error_logger
from sanic.log import logger as root_logger
from sanic.log import server_logger


async def prepare_logger(app: Sanic, *_):
    Logger.prepare(app)


async def setup_logger(app: Sanic, *_):
    logger = Logger()
    app.manager.manage(
        "Logger",
        logger,
        {
            "queue": app.shared_ctx.logger_queue,
        },
    )


class SanicQueueHandler(QueueHandler):
    def emit(self, record: LogRecord) -> None:
        try:
            return super().enqueue(record)
        except Full:
            server_logger.warning(
                "Background logger is full. Emitting log in process."
            )
            server_logger.handle(record)


async def setup_server_logging(app: Sanic):
    qhandler = SanicQueueHandler(app.shared_ctx.logger_queue)
    app.ctx._logger_handlers = defaultdict(list)
    app.ctx._qhandler = qhandler

    for logger_instance in (root_logger, access_logger, error_logger):
        for handler in logger_instance.handlers:
            logger_instance.removeHandler(handler)
        logger_instance.addHandler(qhandler)


async def remove_server_logging(app: Sanic):
    for logger, handlers in app.ctx._logger_handlers.items():
        logger.removeHandler(app.ctx._qhandler)
        for handler in handlers:
            logger.addHandler(handler)


class Logger:
    def __init__(self):
        self.run = True
        self.loggers = {
            logger.name: logger
            for logger in (root_logger, access_logger, error_logger)
        }

    def __call__(self, queue) -> None:
        signal_func(SIGINT, self.stop)
        signal_func(SIGTERM, self.stop)

        while self.run:
            try:
                record: LogRecord = queue.get_nowait()
            except Empty:
                continue
            logger = self.loggers.get(record.name)
            logger.handle(record)

    def stop(self, *_):
        if self.run:
            self.run = False

    @classmethod
    def prepare(cls, app: Sanic):
        sync_manager = Manager()
        logger_queue = sync_manager.Queue(
            maxsize=app.config.LOGGING_QUEUE_MAX_SIZE
        )
        app.shared_ctx.logger_queue = logger_queue

    @classmethod
    def setup(cls, app: Sanic):
        app.main_process_start(prepare_logger)
        app.main_process_ready(setup_logger)
        app.before_server_start(setup_server_logging)
        app.before_server_stop(remove_server_logging)



from functools import partial
from inspect import isawaitable, isclass

from sanic.log import logger

from sanic_ext.exceptions import ValidationError
from sanic_ext.utils.typing import is_pydantic

from .schema import make_schema
from .validators import (
    _validate_annotations,
    _validate_instance,
    validate_body,
)


async def do_validation(
    *,
    model,
    data,
    schema,
    request,
    kwargs,
    body_argument,
    allow_multiple,
    allow_coerce,
):
    try:
        logger.debug(f"Validating {request.path} using {model}")
        if model is not None:
            if isclass(model):
                validator = _get_validator(
                    model, schema, allow_multiple, allow_coerce
                )
                validation = validate_body(validator, model, data)
                kwargs[body_argument] = validation
            else:
                validation = model(
                    request=request, data=data, handler_kwargs=kwargs
                )
                if isawaitable(validation):
                    await validation
    except TypeError as e:
        raise ValidationError(e)


def generate_schema(param):
    try:
        if param is None or is_pydantic(param):
            return param
    except TypeError:
        ...

    return make_schema({}, param) if isclass(param) else param


def _get_validator(model, schema, allow_multiple, allow_coerce):
    if is_pydantic(model):
        return _validate_instance

    return partial(
        _validate_annotations,
        schema=schema,
        allow_multiple=allow_multiple,
        allow_coerce=allow_coerce,
    )



from sys import version_info
from typing import List

import pytest

from sanic_ext.extensions.openapi.types import Schema


@pytest.mark.skipif(version_info < (3, 9), reason="Not needed on 3.8")
def test_schema_list():
    class Foo:
        list1: List[int]
        list2: list[int]

        @property
        def show(self) -> bool:
            return True

        def no_show_method(self) -> None:
            ...

        @classmethod
        def no_show_classmethod(self) -> None:
            ...

        @staticmethod
        def no_show_staticmethod() -> None:
            ...

    schema = Schema.make(Foo)
    serialized = schema.serialize()
    assert "no_show_method" not in serialized
    assert "no_show_classmethod" not in serialized
    assert "no_show_staticmethod" not in serialized
    assert serialized == {
        "type": "object",
        "properties": {
            "list1": {
                "type": "array",
                "items": {"type": "integer", "format": "int32"},
            },
            "list2": {
                "type": "array",
                "items": {"type": "integer", "format": "int32"},
            },
            "show": {"type": "boolean"},
        },
    }



from sanic import Request, Sanic, text

from sanic_ext.extensions.openapi import openapi

from .utils import get_spec


def test_tag(app: Sanic):
    TAG_NAME = "test-tag-1"

    @app.route("/test0")
    async def handler0(request: Request, val1: int):
        """
        openapi:
        ---
        tags:
          - test-tag-1
        """
        return text("ok")

    @app.route("/test1")
    @openapi.tag(TAG_NAME)
    async def handler1(request: Request, val1: int):
        return text("ok")

    @app.route("/test2")
    @openapi.definition(tag=TAG_NAME)
    async def handler2(request: Request, val1: int):
        return text("ok")

    spec = get_spec(app)
    for i in range(3):
        assert f"/test{i}" in spec["paths"]
        tag = spec["paths"][f"/test{i}"]["get"]["tags"][0]
        assert tag == TAG_NAME



from sanic_ext.extensions.openapi.autodoc import YamlStyleParametersParser

tests = []

_ = ""

tests.append({"doc": _, "expects": {}})

_ = "one line docstring"

tests.append({"doc": _, "expects": {"summary": "one line docstring"}})

_ = """
first line

more lines
"""

tests.append(
    {
        "doc": _,
        "expects": {"summary": "first line", "description": "more lines"},
    }
)


_ = """
first line

more lines

openapi:
---
responses:
  '200':
    description: OK
"""

tests.append(
    {
        "doc": _,
        "expects": {
            "summary": "first line",
            "description": "more lines",
            "responses": {"200": {"description": "OK"}},
        },
    }
)


def test_autodoc():
    for t in tests:
        parser = YamlStyleParametersParser(t["doc"])
        assert parser.to_openAPI_2() == t["expects"]
        assert parser.to_openAPI_3() == t["expects"]



from sanic import text
from sanic.response import json

from sanic_ext import serializer


def test_serializer_with_builtin(app):
    @app.get("/")
    @serializer(text)
    async def handler(request):
        return "hello"

    @app.get("/201")
    @serializer(text, status=201)
    async def handler_201(request):
        return "hello"

    _, response = app.test_client.get("/")
    assert response.status_code == 200
    assert response.text == "hello"
    assert response.content_type == "text/plain; charset=utf-8"

    _, response = app.test_client.get("/201")
    assert response.status_code == 201
    assert response.content_type == "text/plain; charset=utf-8"


def test_serializer_with_custom(app):
    def custom(message, status):
        return json({"message": message}, status=status)

    @app.get("/")
    @serializer(custom)
    async def handler(request):
        return "hello"

    @app.get("/201")
    @serializer(custom, status=201)
    async def handler_201(request):
        return "hello"

    _, response = app.test_client.get("/")
    assert response.status_code == 200
    assert response.content_type == "application/json"
    assert response.json["message"] == "hello"

    _, response = app.test_client.get("/201")
    assert response.status_code == 201
    assert response.content_type == "application/json"


def test_serializer_with_params(app):
    def message(retval, request, action, status):
        return json(
            {
                "request_id": str(request.id),
                "action": action,
                "message": retval,
            },
            status=status,
        )

    @app.get("/<action>")
    @serializer(message)
    async def do_action(request, action: str):
        return "This is a message"

    _, response = app.test_client.get("/this")
    assert response.status_code == 200
    assert response.json["action"] == "this"



import types
import typing

try:
    UnionType = types.UnionType  # type: ignore
except AttributeError:
    UnionType = type("UnionType", (), {})

try:
    from pydantic import BaseModel

    PYDANTIC = True
except ImportError:
    PYDANTIC = False

try:
    import attrs  # noqa

    ATTRS = True
except ImportError:
    ATTRS = False


def is_generic(item):
    return (
        isinstance(item, typing._GenericAlias)
        or isinstance(item, UnionType)
        or hasattr(item, "__origin__")
    )


def is_optional(item):
    if is_generic(item):
        args = typing.get_args(item)
        return len(args) == 2 and type(None) in args
    return False


def is_pydantic(model):
    return PYDANTIC and (
        issubclass(model, BaseModel) or hasattr(model, "__pydantic_model__")
    )


def is_attrs(model):
    return ATTRS and (hasattr(model, "__attrs_attrs__"))
